{-

  APCC
  
  Copyright Paolo Modesti, School of Computing Science, Newcastle University 2015

-}

module StrandMP where
import Ast
import Msg
import TypeSystem_Defs
import StrandAST
import GeneratePlainStrands (generateStrands)
import OperationalStrands (opStrands)
import MessagePattern
import Control.Monad.State
import Unification
import Data.List
import Debug.Trace


type Unification = ((Term,String),(Term,String))
type Unifications = [Unification]

eqUni :: Unification -> Unification -> Bool
eqUni ((t1,p1),(t2,p2)) ((t3,p3),(t4,p4)) = (t1==t4) && (t2==t3) && (p1==p4) && (p2==p3)

showUni :: Unifications -> String
showUni [] = ""
showUni (((t1,p1),(t2,p2)):xs) = ("\n------- Unification between terms:\nProtocol: "++ p1 ++"\nTerm: " ++ show t1 ++ "\nTerm: " ++ show t2 ++ "\nProtocol: " ++ p2) ++ showUni xs

unifyParijU :: [(Term,String)] -> [(Term,String)] -> Unifications
unifyParijU [] [] = []
unifyParijU _ [] = []
unifyParijU [] _ = []
unifyParijU (x:xs) y = unifyAuxParU x y ++ unifyParijU xs y

unifyAuxParU :: (Term, String) -> [(Term,String)] -> Unifications
unifyAuxParU (t1,p1) array =  foldl (\acc (t2,p2) -> if unify [(t1, t2)] /= Nothing 
                                                                then acc ++ [((t1, p1), (t2, p2))]
                                                                -- then [((t1, p1), (t2, p2))]         -- less unifications but still significant 
                                                                else acc) [] array


smpDisjointCheckerU :: [Protocol] -> Unifications
smpDisjointCheckerU [] = []
smpDisjointCheckerU (x:arrayTail) = let 
                                         uni = foldl (\acc a -> smpDisjointCheckerijU x a ++ acc ) [] arrayTail ++ smpDisjointCheckerU arrayTail
                                    in nubBy eqUni uni 

smpDisjointCheckerijU :: Protocol -> Protocol -> Unifications
smpDisjointCheckerijU p1 p2 = let 
                                smp1 = ttsSMP p1
                                smp2 = ttsSMP p2
                                smp3 = unifyParijU smp1 smp2
                             in smp3 
showttsSMP :: Protocol -> String
showttsSMP prot = ppSMP . ttsSMP $ prot

ttsSMP :: Protocol -> [(Term,String)]
-- ttsSMP prot@(name,_,_,_,_,_,_,_,_,_,_) | trace ("ttsSMP\tProtocol: " ++ show name) False = undefined
ttsSMP prot@(name,_,_,_,_,_,_,_,_,_,_) = let 
                                            os = opStrands . generateStrands $prot
                                            (smp1,_) = evalState (mpStrandList name os) ([], 0)
                                            (smp2,_) = protocolNVST smp1 ([],0)
                                            smp3 = sort (filter notisPublic smp2)
                                         in smp3 -- error(show smp3) 

notisPublic :: (Term,String) -> Bool
notisPublic (AMsg (Comp Pub _ _), _) = False
notisPublic _ = True

showSMP :: StateType -> String 
showSMP (smp,_) = ppSMP smp

ppSMP :: [(Term, String)] -> String
ppSMP [] = ""
ppSMP ((t, _):xs) = show t ++ "\n" ++ ppSMP xs

mpStrand :: String -> Strand -> State StateType StateType
mpStrand protName (Strand n o k r) = mpStrand0 protName r

mpStrandList :: String -> [Strand] -> State StateType StateType
mpStrandList protName xs = do
                                      mapM_ (mpStrand protName) xs
                                      get

strPhiRest :: Strand0 -> ([(Term,Term)],Strand0)
strPhiRest (How d (Atom id) m2 m3 rest) = ((AMsg m2, AMsg m3) : ts,str) where
                                                  (ts,str) = strPhiRest rest                      

strPhiRest (V v m1 m2 rest) = ((AMsg m1, AMsg m2) : ts,str) where
                                   (ts,str) = strPhiRest rest
strPhiRest str = ([],str)

appSubsTerm :: Term -> [(Term,Term)] -> Term
appSubsTerm = foldr (\ (t1, t2) x -> substitution t1 t2 x)

appSubsStrand0 :: Strand0 -> [(Term,Term)] -> Strand0
appSubsStrand0 (Snd det m rest) subs = Snd det m1 rest1 
                                        where 
                                                rest1 = appSubsStrand0 rest subs
                                                (AMsg m1) = appSubsTerm (AMsg m) subs
appSubsStrand0 (Rcv det m rest) subs = Rcv det m1 rest1 
                                        where 
                                                rest1 = appSubsStrand0 rest subs
                                                (AMsg m1) = appSubsTerm (AMsg m) subs
-- OR and IF not supported yet 
appSubsStrand0 (OR lb rb rest) subs = OR lb rb rest1 
                                        where 
                                                rest1 = appSubsStrand0 rest subs
appSubsStrand0 (IF fct tb fb rest) subs = IF fct tb fb rest1 
                                        where 
                                                rest1 = appSubsStrand0 rest subs

appSubsStrand0 (Add fct s m ms rest) subs = Add fct s m1 ms rest1
                                        where 
                                                rest1 = appSubsStrand0 rest subs
                                                (AMsg m1) = appSubsTerm (AMsg m) subs
appSubsStrand0 (Delete fct rest) subs = Delete fct rest1
                                        where 
                                                rest1 = appSubsStrand0 rest subs
appSubsStrand0 (Query fct rest) subs = Query fct rest1
                                        where 
                                                rest1 = appSubsStrand0 rest subs
appSubsStrand0 (Omar str rest) subs = Omar str rest1
                                        where 
                                                rest1 = appSubsStrand0 rest subs
appSubsStrand0 (Connects rest) subs = Connects rest1
                                        where 
                                                rest1 = appSubsStrand0 rest subs
appSubsStrand0 (How d m1 m2 m3 rest) subs = How d ma1 ma2 ma3 rest1
                                        where 
                                                rest1 = appSubsStrand0 rest subs
                                                (AMsg ma1) = appSubsTerm (AMsg m1) subs
                                                (AMsg ma2) = appSubsTerm (AMsg m2) subs
                                                (AMsg ma3) = appSubsTerm (AMsg m3) subs
appSubsStrand0 (V v m1 m2  rest) subs = V v ma1 ma2 rest1
                                        where 
                                                rest1 = appSubsStrand0 rest subs
                                                (AMsg ma1) = appSubsTerm (AMsg m1) subs
                                                (AMsg ma2) = appSubsTerm (AMsg m2) subs
appSubsStrand0 (Create m1 m2 m3 rest) subs = Create ma1 ma2 ma3 rest1
                                        where 
                                                rest1 = appSubsStrand0 rest subs
                                                (AMsg ma1) = appSubsTerm (AMsg m1) subs
                                                (AMsg ma2) = appSubsTerm (AMsg m2) subs
                                                (AMsg ma3) = appSubsTerm (AMsg m3) subs
appSubsStrand0 (Fresh t ms m3 rest) subs = Fresh t ms m3 rest1 
                                        where 
                                                rest1 = appSubsStrand0 rest subs
appSubsStrand0 str@(EOS _) _ = str 



mpStrand0 :: String -> Strand0 -> State StateType StateType
mpStrand0 protName (Snd _ m rest) = do 
                                           msgPattern (AMsg m,protName)
                                           mpStrand0 protName rest
                                           get 

mpStrand0 protName (Rcv (_,expr,_,_) (Atom id) rest) = 
                                            let 
                                                (ts,rest1) = strPhiRest rest
                                                -- compute the mgu
                                                subs = unify ts
                                            in -- make a substitution mgu of 
                                               case subs of
                                                    Nothing -> do  
                                                        -- no substitution
                                                        -- compute msg patterns (and alpharenaming)
                                                        msgPattern (AMsg expr,protName)
                                                        mpStrand0 protName rest1
                                                        get 
                                                    Just subst -> let 
                                                                    --- Substitutions and 
                                                                    rest2 = appSubsStrand0 rest1 subst
                                                                    msg1 = appSubsTerm (AMsg expr) subst 
                                                                  in do  
                                                                    -- alpharenaming 
                                                                    msgPattern (msg1,protName)
                                                                    mpStrand0 protName rest2
                                                                    get
                                           
mpStrand0 protName (Add fct s m ms rest) = do 
                                           msgPattern (AMsg m,protName)
                                           mpStrand0 protName rest
                                           get 
                                           
mpStrand0 protName (Fresh t ms@(m@(Atom id): _) m3 rest) = do 
                                           mpStrand0 protName rest
                                           get                                            

                                          
mpStrand0 protName (How d (Atom id) m2 m3 rest) = do 
                                           mpStrand0 protName rest
                                           get 

mpStrand0 protName (V Check m1 m2 rest) = do 
                                           msgPattern (AMsg m1,protName)
                                           msgPattern (AMsg m2,protName)
                                           mpStrand0 protName rest
                                           get 

mpStrand0 protName (V v m1 m2 rest) = do 
                                           mpStrand0 protName rest
                                           get 

---- just MP the rest
--IF and OR not implemented yet                                           
mpStrand0 protName (IF _ _ _ rest) = mpStrand0 protName rest
mpStrand0 protName (OR _ _ rest) = mpStrand0 protName rest 
mpStrand0 protName (Delete _ rest) = mpStrand0 protName rest 
mpStrand0 protName (Query _ rest) = mpStrand0 protName rest
mpStrand0 protName (Omar _ rest) = mpStrand0 protName rest                                           
mpStrand0 protName (Connects rest) = mpStrand0 protName rest 
mpStrand0 protName (Create _ _ _ rest) = mpStrand0 protName rest
mpStrand0 _ (EOS _) = get

mpStrand0 _ str = error ("unexpected strand type: " ++ show str)