{-

  APCC  

  Copyright Paolo Modesti, School of Computing Science, Newcastle University 2015 

-}

{-# LANGUAGE DeriveDataTypeable #-}

module TypeSystem_Defs where

import Data.Typeable
import Data.Data
import Ast (FormatName,MappingName)
import Msg (Msg,Ident)
import StrandAST (Destructor,Verifier)

data TermType = TBoolean
             | TNumber
             | TAgent
             | TPublicKey
             | TPrivateKey
             | TSymmetricKey
             | TImpData
             | TScrypt (TermType,TermType)
             | TCrypt (TermType,TermType)
             | TSign (TermType,TermType)
             | TExp (TermType,TermType)
             | TMult (TermType,TermType)
             | THash TermType   
             | TMac (TermType,TermType) 
       --      | TFunction Ident ([TermType],TermType)
             | TFunction Ident [TermType]
             | TFormat FormatName [TermType]
             | TMapping MappingName ([TermType],TermType)
             | TCat [TermType] -- actually not allowed in APS
             | TSeqNumber
             | TUntyped
             | TMsg
      deriving (Eq,Typeable,Data,Show,Ord) 


-- here wee define a specific function for comparing types, used for typechecking
-- instance Eq cannot be useded because we need to consieder formats with Msg

compareTypes :: TermType -> TermType -> Bool 
compareTypes (TFormat id1 xs1) (TFormat id2 xs2) = id1==id2 && length xs1 == length xs2 && and (zipWith compareTFormatsTypes xs1 xs2) 
compareTypes t1 t2 = t1==t2

compareTFormatsTypes :: TermType -> TermType -> Bool 
compareTFormatsTypes TMsg _ = True
compareTFormatsTypes _ TMsg = True
compareTFormatsTypes t1 t2 = t1==t2 

data Term =   AMsg Msg
            | TmVar String Binding
            | TmBind String Binding -- a binding 
            | ADes Destructor Msg Msg 
            | AVer Verifier Msg Msg    
          deriving (Eq,Typeable,Data,Show,Ord) 

-- binding
     
data Binding = VarBind TermType
               deriving (Eq,Typeable,Data,Show,Ord)
