{-

  APCC
  
  Copyright Paolo Modesti, School of Computing Science, Newcastle University 2015
  Copyright Ivana Kojovic, DTU, 2012 
  Minor Update by Omar Almousa 2015

-}

module Unification where
import TypeSystem_Defs
import Msg

term2msg :: Term -> Msg 
term2msg t = case t of
                AMsg a -> a
                _ -> error ("term2msg error: " ++ show t)     

-- | Substitution of the Terms on single term
substitution :: Term -> Term -> Term -> Term
substitution (AMsg (Atom x)) t (AMsg (Atom s))  | s == x    = t      
                                                | otherwise = AMsg (Atom s)

-- | Substitution of the Terms on complex (function) term
-- substitution x t (Atom (Mapping id a)) = Atom (Mapping id (map toAtomic (map (substitution x t) (wrapIdent a)))) -- TODO do we need something special for mappings?

substitution x t  (AMsg (Comp op o xs)) = AMsg (Comp op o ms) 
                                                    where 
                                                       t1 = map (substitution x t . AMsg) xs
                                                       ms = map term2msg t1
                                                       
substitution x t (ADes dec m1 m2) = ADes dec msg1 msg2  
                                                where 
                                                   msg1 = term2msg (substitution x (AMsg m1) t)
                                                   msg2 = term2msg (substitution x (AMsg m2) t)

substitution x t (AVer dec m1 m2) = AVer dec msg1 msg2  
                                                where 
                                                   msg1 = term2msg (substitution x (AMsg m1) t)
                                                   msg2 = term2msg (substitution x (AMsg m2) t)
                                                   

substitution (TmVar id (VarBind ty)) t t1@(TmVar id1 (VarBind ty1)) | id == id1 && ty==ty1 = t      
                                                                    | otherwise = t1

substitution (TmBind id (VarBind ty)) t t1@(TmVar id1 (VarBind ty1)) | id == id1 && ty==ty1 = t      
                                                                     | otherwise = t1


substitution x _ _ = error("substitution non possible in " ++ show x)

substitutionp x t (p1,p2)    = (substitution x t p1, substitution x t p2)


-- | if variable is "inside" the non atomic term on the right side -}

occursCheck:: (String, Term) -> Bool

occursCheck (x, AMsg(Atom t)) | x==t = True 
                              | otherwise = False
occursCheck (x, AMsg (Comp op o xs)) = foldl (\acc a -> (occursCheck (x, a) || acc)) False (map AMsg xs)
occursCheck (x, ADes dec m1 m2) = occursCheck (x,AMsg m1) || occursCheck (x,AMsg m2)  
occursCheck (x, AVer ver m1 m2) = occursCheck (x,AMsg m1) || occursCheck (x,AMsg m2)
occursCheck (x, TmVar id (VarBind ty)) | x==id = True 
                                       | otherwise = False
occursCheck (x, TmBind id (VarBind ty)) | x==id = True 
                                        | otherwise = False


---------------------------------   Wrap Term  ---------technical stuff, not so important

--wrapIdent :: [Atomic] -> [Term]
--wrapIdent [] = []
--wrapIdent (x:xs) = (Atom x): wrapIdent xs

 --------------------------------- Making atomic Term ---- also not so important

--toAtomic :: Term -> Atomic
--toAtomic (Atom (Ident t)) = Ident t
--toAtomic (Atom (Mapping id t)) = Mapping id t


isVarTerm :: Term -> Bool -- 
isVarTerm (AMsg (Atom id)) = isVariable id
isVarTerm (TmVar id _) = isVariable id
isVarTerm _ = False

-- | Unification algorithm also works for set of the Terms

unify :: [(Term, Term)] -> Maybe [(Term, Term)]

unify xx = unify' xx []

unify' :: [(Term, Term)] -> [(Term, Term)] -> Maybe [(Term, Term)]

-- | When there is nothing to process, acc is the solution
unify' [] acc = Just acc

-- | Delete rule
unify' ((a, b):xx) acc | a==b = unify' xx acc


-- | Substitute rule : takes care of occurrence of the variable of on the right side; constants on the left side => do the substitution with switched arguments

unify' ((AMsg (Atom x),t):xx) acc | not (occursCheck (x,t)) && isVariable x = -- SUBSTITUTE - watching out on constants!
                                    unify' (map (substitutionp (AMsg (Atom x)) t) xx) ((AMsg (Atom x),t) : map (substitutionp (AMsg (Atom x)) t) acc)
                                  | isVarTerm t && not (isVariable x)  =  
                                          unify' (map (substitutionp t (AMsg (Atom x))) xx) ((t, AMsg (Atom x)) : map (substitutionp t  (AMsg (Atom x))) acc)

                                  | otherwise = Nothing

-- | Orient rule
unify' ((t,AMsg (Atom x)):xx) acc = unify' ((AMsg (Atom x), t):xx) acc 

-- | Decompose rules for all the non atomic Term terms, including mappings!

unify' ((AMsg (Comp op1 o1 xs1),AMsg (Comp op2 o2 xs2)):xx) acc 
                                | op1==op2 && length xs1 == length xs2 = unify' (zip (map AMsg xs1) (map AMsg xs2) ++ xx) acc
                                | otherwise = Nothing 

unify' ((ADes dec1 m11 m12, ADes dec2 m21 m22):xx) acc 
                                        | dec1==dec2 = unify' ([(AMsg m11,AMsg m21)]++([(AMsg m12,AMsg m22)] ++ xx)) acc
                                        | otherwise = Nothing 

unify' ((AVer ver1 m11 m12, AVer ver2 m21 m22):xx) acc 
                                        | ver1==ver2 = unify' ([(AMsg m11,AMsg m21)]++([(AMsg m12,AMsg m22)] ++ xx)) acc
                                        | otherwise = Nothing 

unify' ((TmVar id2 (VarBind t1), TmVar id1 (VarBind t2)):xx) acc 
                                        | id1==id2 && t1==t2 = unify' ((AMsg (Atom id1), AMsg (Atom id2)) : xx) acc
                                        | otherwise = Nothing 
-- 
unify' ((t,TmVar id (VarBind ty)):xx) acc = unify' ((TmVar id (VarBind ty), t):xx) acc 


-- | For everything else, return Nothing
unify' (_) _ = Nothing  

{- ----------------------------------------------- test-------------------------------------------------------------- -}
